#include <proto.h>

#include "gtscroller.h"
#include "gtknob.h"
#include "gtpulseknob.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtScroller class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init scroller.
*
\************************************************************************/
GtScroller *GtMkScroller(char *n,GtManager *p,GtTag *tgs)
{
	GtScroller *obj;
	
	if(!(obj=new GtScroller()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtScroller object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init scroller.
*
\************************************************************************/
GtScroller *GtMkScroller(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtScroller *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkScroller(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Setup empty scroller.
*
\************************************************************************/
GtScroller::GtScroller(void)
{
	maxchild=1;
	knob=NULL;
	start=0;
	range=100;
	page=1;
	val=0;
	knpos=0;
	knsz=0;
	kmxs=1;
	kmys=1;
	orient=GTO_HORIZ;
	sflags=0;
	ibw=0;
	bdrw=1;
	ibdrw=0;
	ftype=GTF_DBL_OUT;
}

/************************************************************************\
*
* Setup scroller.
*
\************************************************************************/
sbool GtScroller::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	char *kname="knob";
	long s=0,e=100,del=0;
	
	if(!GtManager::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_ORIENT,GTRT_ORIENT,&orient);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				orient=t->d.val;
				if(orient>GTO_LAST_NR)
					orient=GTO_HORIZ;
				break;
			case GT_Start:
				s=(long)t->d.val;
				break;
			case GT_End:
				e=(long)t->d.val;
				break;
			case GT_Value:
				val=t->d.val;
				break;
			case GT_Page:
				page=t->d.val;
				break;
			case GT_ReportDrag:
				if(t->d.val)
					sflags|=GTSF_REPDRAG;
				else
					sflags&=~GTSF_REPDRAG;
				break;
			case GT_ReduceInterval:
				if(t->d.val)
					sflags|=GTSF_REDUCE;
				else
					sflags&=~GTSF_REDUCE;
				break;
			case GT_KnobName:
				kname=(char *)t->d.ptr;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				break;
			case GT_Delay:
				del=t->d.val;
				break;
			case GT_Pulse:
				if(t->d.val)
					sflags|=GTSF_PULSE;
				else
					sflags&=~GTSF_PULSE;
				break;
		}
	}
	switch(ftype)
	{
		case GTF_NONE:
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=ibdrw+bdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=ibdrw+2*bdrw;
			break;
	}
	if(sflags & GTSF_PULSE)
	{
		if((s>0) && (e>0))
			s=0;
		if((s<0) && (e<0))
			e=0;
	}
	start=GTMIN(s,e);
	e=e-s;
	if(e<0)
		range=-e;
	else
		range=e;
	if(page>range)
		page=range;
	if(sflags & GTSF_REDUCE)
		rrange=range-page;
	else
		rrange=range;
	if(sflags & GTSF_PULSE)
		val=0;
	else
		val=GTMAX(start,GTMIN(start+(long)rrange,val));
	if(kname)
	{
		if(sflags & GTSF_PULSE)
		{		
			if(!(knob=GtMkPulseKnob(kname,this,
					GT_Disabled,(flags & GTWF_DISABLED ? TRUE : FALSE),
					GT_Delay,del,
					GT_DONE)))
				return(FALSE);
		}
		else
		{
			if(!(knob=GtMkKnob(kname,this,
					GT_Disabled,(flags & GTWF_DISABLED ? TRUE : FALSE),
					GT_DONE)))
				return(FALSE);
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Set scroller attributes.
*
\************************************************************************/
sbool GtScroller::Set(GtTag *tgs)
{
	GtTag *t;
	long s,e;
	sbool och=FALSE,ch=FALSE,rch=FALSE,vch=FALSE;

	if(!GtManager::Set(tgs))
		return(FALSE);
	s=start;
	e=start+range;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				orient=t->d.val;
				if(orient>GTO_LAST_NR)
					orient=GTO_HORIZ;
				och=TRUE;
				break;
			case GT_Start:
				s=t->d.val;
				rch=TRUE;
				break;
			case GT_End:
				e=t->d.val;
				rch=TRUE;
				break;
			case GT_Value:
				val=t->d.val;
				ch=TRUE;
				break;
			case GT_Page:
				page=t->d.val;
				rch=TRUE;
				break;
			case GT_ReportDrag:
				if(t->d.val)
					sflags|=GTSF_REPDRAG;
				else
					sflags&=~GTSF_REPDRAG;
				break;
			case GT_ReduceInterval:
				if(t->d.val)
					sflags|=GTSF_REDUCE;
				else
					sflags&=~GTSF_REDUCE;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				vch=TRUE;
				break;
			case GT_Delay:
				if(knob)
					knob->SetV(GT_Delay,t->d.val,GT_DONE);
				break;
		}
	}
	if(vch)
	{
		switch(ftype)
		{
			case GTF_NONE:
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=ibdrw+bdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=ibdrw+2*bdrw;
				break;
		}
	}
	if(rch || ch)
	{
		if(rch)
		{
			if(sflags & GTSF_PULSE)
			{
				if((s>0) && (e>0))
					s=0;
				if((s<0) && (e<0))
					e=0;
			}
			start=GTMIN(s,e);
			e=e-s;
			if(e<0)
				range=-e;
			else
				range=e;
			if(page>range)
				page=range;
			if(sflags & GTSF_REDUCE)
				rrange=range-page;
			else
				rrange=range;
		}
		if(sflags & GTSF_PULSE)
			val=0;
		else
			val=GTMAX(start,GTMIN(start+(long)rrange,val));
	}
	if(ch || och || vch)
	{
		this->Relayout();
	}
	return(TRUE);
}

/************************************************************************\
*
* Get scroller attributes.
*
\************************************************************************/
void GtScroller::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtManager::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Orient:
				*ptr=orient;
				break;
			case GT_Start:
				*ptr=start;
				break;
			case GT_End:
				*ptr=start+range;
				break;
			case GT_Value:
				*ptr=val;
				break;
			case GT_Page:
				*ptr=page;
				break;
			case GT_ReportDrag:
				if(sflags & GTSF_REPDRAG)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_ReduceInterval:
				if(sflags & GTSF_REDUCE)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_Delay:
				if(knob)
					knob->GetV(GT_Delay,ptr,GT_DONE);
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtScroller::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtScroller")==0)
		return(TRUE);
	else
		return(GtManager::IsSubClassOf(cn));
}

/************************************************************************\
*
* Allow only subclasses of GtKnob to be added.
*
\************************************************************************/
sbool GtScroller::AddChild(GtWidget *o,ulong pos)
{
	if(o && o->Manage())
	{
		if(o->IsSubClassOf("GtKnob"))
		{
			knob=(GtKnob *)o;
			if(knob->IsSubClassOf("GtPulseKnob"))
				sflags|=GTSF_PULSE;
			return(GtManager::AddChild(o,pos));
		}
		else
		{
			GtError(this,GTEC_NOADD,"Unable to add non-knob child: %s\n",o->GetName());
			return(FALSE);
		}
	}
	else
		return(GtManager::AddChild(o,pos));
}

/************************************************************************\
*
* Child asked to be moved and/or resized.
*
\************************************************************************/
void GtScroller::MoveSizeReq(GtWidget *obj,word xp,word yp,word xs,word ys)
{
	GtScrollerCD cd;
	word tsz,oxs,oys;
	long v;

	if(flags & GTWF_UNBUILT)
	{
		obj->SetPosSize(xp,yp,xs,ys);
		return;
	}
	oxs=xs;
	oys=ys;
	v=val;
	if(orient==GTO_HORIZ)
	{
		if(xp<ibw)
			xp=ibw;
		if(xp+knsz>xsize-ibw)
			xp=xsize-knsz-ibw;
		knpos=xp;
		yp=ibw;
		xs=knsz;
		ys=ysize-2*ibw;
		tsz=xsize;
	}
	else
	{
		if(yp<ibw)
			yp=ibw;
		if(yp+knsz>ysize-ibw)
			yp=ysize-knsz-ibw;
		knpos=yp;
		xp=ibw;
		xs=xsize-2*ibw;
		ys=knsz;
		tsz=ysize;
	}
	if(tsz-knsz-2*ibw)
		val=start+rrange*(knpos-ibw)/(tsz-knsz-2*ibw);
	if(knpos==ibw)
		val=start;
	if(knpos+knsz==tsz-ibw+1)
		val=start+rrange;
	val=GTMAX(start,GTMIN(start+(long)rrange,val));
	if((oxs>kmxs) || (oys>kmys))
		this->Relayout();
	else
		obj->SetPosSize(xp,yp,xs,ys);
	if((sflags & GTSF_REPDRAG) && (v!=val))
	{
		cd.value=val;
		cd.drag=TRUE;
		chg.CallCB(this,&cd);
	}
}

/************************************************************************\
*
* Relayout scroller knob.
*
\************************************************************************/
void GtScroller::Relayout(void)
{
	int xs=1,ys=1;
	
	if(flags & GTWF_UNBUILT)
		return;
	if(knob)
		knob->GetMinSize(&xs,&ys);
	kmxs=GTMAX(xs,kmxs);
	kmys=GTMAX(ys,kmys);
	this->SetMinMaxSize(GTMAX(2*ibw+kmxs,2*ibw+1),GTMAX(2*ibw+kmys,2*ibw+1),
											GTS_MAX,GTS_MAX);
}

/************************************************************************\
*
* Set scroller event mask.
*
\************************************************************************/
void GtScroller::SetEMask(ulong em)
{
	GtManager::SetEMask(em|ExposureMask|ButtonReleaseMask|ButtonPressMask);
}

/************************************************************************\
*
* Add callback.
*
\************************************************************************/
void GtScroller::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	if(tag==GT_ValueChange)
		chg.AddCB(cb,ud);
	else
		GtManager::AddCB(tag,cb,ud);
}

/************************************************************************\
*
* Remove callback.
*
\************************************************************************/
void GtScroller::RemCB(ulong tag,GtCallB cb)
{
	if(tag==GT_ValueChange)
		chg.RemCB(cb);
	else
		GtManager::RemCB(tag,cb);
}

/************************************************************************\
*
* Notify from knob release.
*
\************************************************************************/
void GtScroller::Notify(ulong tag)
{
	GtScrollerCD cd;
	
	if(tag==GT_Activate)
	{
		cd.value=val;
		cd.drag=FALSE;
		chg.CallCB(this,&cd);
	}
	else
		GtManager::Notify(tag);
}

/************************************************************************\
*
* Layout scroller and knob.
*
\************************************************************************/
void GtScroller::Layout(void)
{
	int xp,yp,xs,ys;

	if(flags & GTWF_UNBUILT)
		return;
	if(knob)
	{
		knob->GetMinSize(&xs,&ys);
		kmxs=GTMAX(xs,kmxs);
		kmys=GTMAX(ys,kmys);
		if(orient==GTO_HORIZ)
		{
			if(range)
				knsz=kmxs+(xsize-2*ibw-kmxs)*page/range;
			else
				knsz=kmxs+(xsize-2*ibw-kmxs);
			knsz=GTMAX(knsz,xs);
			if(rrange)
				knpos=ibw+(xsize-2*ibw-knsz)*(val-start)/rrange;
			else
				knpos=ibw;
			xp=knpos;
			yp=ibw;
			xs=knsz;
			ys=ysize-2*ibw;
		}
		else
		{
			if(range)
				knsz=kmys+(ysize-2*ibw-kmys)*page/range;
			else
				knsz=kmys+(ysize-2*ibw-kmys);
			knsz=GTMAX(knsz,ys);
			if(rrange)
				knpos=ibw+(ysize-2*ibw-knsz)*(val-start)/rrange;
			else
				knpos=ibw;
			yp=knpos;
			xp=ibw;
			ys=knsz;
			xs=xsize-2*ibw;
		}
		knob->SetPosSize(xp,yp,xs,ys);
	}
}

/************************************************************************\
*
* Draw scroller.
*
\************************************************************************/
sbool GtScroller::Draw(XEvent *xe)
{
	int n;

	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
// Draw disabled
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,n,n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;			
	}
	return(TRUE);
}

/************************************************************************\
*
* Mouse scroller events.
*
\************************************************************************/
sbool GtScroller::Mouse(XEvent *xe)
{
	GtScrollerCD cd;
	long xp,yp,v;

	if(flags & GTWF_DISABLED)
		return(FALSE);
	if((xe->type==ButtonRelease) && (xe->xbutton.button==Button1))
	{
		if(sflags & GTSF_PULSE)
			return(TRUE);
		xp=xe->xbutton.x;
		yp=xe->xbutton.y;
		if((xp>ibw) && (xp<xsize-ibw) && (yp>ibw) && (yp<ysize-ibw))
		{
			v=val;
			if(orient==GTO_HORIZ)
			{
				if(xp<knpos)
					val-=page;
				else
					val+=page;
			}
			else
			{
				if(yp<knpos)
					val-=page;
				else
					val+=page;
			}
			val=GTMAX(start,GTMIN(start+(long)rrange,val));
			this->Layout();
			if(v!=val)
			{
				cd.value=val;
				cd.drag=FALSE;
				chg.CallCB(this,&cd);
			}
		}
		return(TRUE);
	}
	return(FALSE);
}
